from setuptools import setup

setup(name='twitter_data_equipment',
      version='1.4.8',
      description='A package that wraps tools for harvesting, hydrating, cleaning and visualize twitter data',
      url='https://github.com/HolzmanoLagrene/FacharbeitLenzBaumannHS16FS17',
      author='Lenz Baumann',
      author_email='lnzbmnn@gmail.com',
      license='MIT',
      packages=['twitter_data_equipment'],
      include_package_data=True,
      zip_safe=False)

